---------------------------------------------------------
-- 1) ROUNDED RECTANGLE MESH CODE
---------------------------------------------------------
rrectshad = {
    vert = [[
    uniform mat4 modelViewProjection;
    attribute vec4 position;
    attribute vec2 texCoord;
    attribute vec3 normal;
    
    varying highp vec2 vTexCoord;
    varying vec3 vNormal;
    
    void main()
    {
    vTexCoord = texCoord;
    vNormal = normal;
    gl_Position = modelViewProjection * position;
    }
    ]],
    
    frag = [[
    precision highp float;
    uniform lowp vec4 fillColor;
    uniform lowp vec4 strokeColor;
    uniform float scale;
    uniform float strokeWidth;
    varying highp vec2 vTexCoord;
    varying vec3 vNormal;
    
    void main()
    {
    // Basic fill/stroke blend
    lowp vec4 col = mix(strokeColor, fillColor,
    smoothstep((1.0 - strokeWidth) - scale * 0.5,
    (1.0 - strokeWidth) - scale * 1.5,
    vNormal.z));
    // Soft alpha on the outer edge
    col = mix(vec4(col.rgb, 0.0),
    col,
    smoothstep(1.0, 1.0 - scale, vNormal.z));
    gl_FragColor = col;
    }
    ]]
}

local __RRects = {}

function roundedRectangle(t)
    local s = t.radius or 8
    local c = t.corners or 15  -- 15 => round all corners
    local w = math.max(t.w+1, 2*s) + 1
    local h = math.max(t.h, 2*s) + 2
    
    local mode = rectMode()
    if mode == CORNER then
        t.x = t.x + t.w/2
        t.y = t.y + t.h/2
    end
    
    local label = table.concat({
        w, h, s, c,
        mode, -- check rectMode
        0, 0
    }, ",")
    
    if not __RRects[label] then
        local rr = mesh()
        rr.shader = shader(rrectshad.vert, rrectshad.frag)
        
        local v, no = {}, {}
        local n = math.max(3, s//2)
        local edge, cent = vec3(0,0,1), vec3(0,0,0)
        
        local function cornerPoly(dx, dy, o)
            for i = 1, n do
                table.insert(v, o)
                table.insert(v, o + vec2(dx*s*math.cos((i-1)*math.pi/(2*n)), dy*s*math.sin((i-1)*math.pi/(2*n))))
                table.insert(v, o + vec2(dx*s*math.cos(i*math.pi/(2*n)),   dy*s*math.sin(i*math.pi/(2*n))))
                
                table.insert(no, cent)
                table.insert(no, edge)
                table.insert(no, edge)
            end
        end
        
        local function cornerBlock(o, dx, dy)
            v[#v+1] = o
            v[#v+1] = o + vec2(dx*s, 0)
            v[#v+1] = o + vec2(dx*s, dy*s)
            
            v[#v+1] = o
            v[#v+1] = o + vec2(0, dy*s)
            v[#v+1] = o + vec2(dx*s, dy*s)
            
            local blockNormals = {cent, edge, edge, cent, edge, edge}
            for i=1,#blockNormals do
                no[#no+1] = blockNormals[i]
            end
        end
        
        -- Build the corners
        for j = 1, 4 do
            local dx = 1 - 2*(((j+1)//2)%2)
            local dy = -1 + 2*((j//2)%2)
            local o = vec2(dx*(w*0.5 - s), dy*(h*0.5 - s))
            local bit = 2^(j-1)
            
            if c & bit == bit then
                cornerPoly(dx, dy, o)
            else
                cornerBlock(o, dx, dy)
            end
        end
        
        rr.vertices = v
        
        -- Now add center rectangles
        rr:addRect(0, 0,         w-2*s, h-2*s)
        rr:addRect(0, (h-s)/2,   w-2*s, s)
        rr:addRect(0, -(h-s)/2,  w-2*s, s)
        rr:addRect(-(w-s)/2, 0,  s, h-2*s)
        rr:addRect((w-s)/2,  0,  s, h-2*s)
        
        -- Normals for center parts
        local centerNormals = {
            cent,cent,cent, cent,cent,cent,
            edge,cent,cent, edge,cent,edge,
            cent,edge,edge, cent,edge,cent,
            edge,edge,cent, edge,cent,cent,
            cent,cent,edge, cent,edge,edge
        }
        for i=1,#centerNormals do
            no[#no+1] = centerNormals[i]
        end
        
        rr.normals = no
        rr.shader.scale = 0.25 / math.max(2, s)
        
        local sc = 0.25/math.max(2, s)
        rr.shader.strokeWidth = math.min( 1 - sc*3, strokeWidth() * sc)
        
        __RRects[label] = rr
    end
    
    -- Respect the current fill/stroke in Codea
    local rr = __RRects[label]
    
    -- Each time we use this mesh, update a timestamp
    rr._lastUsed = ElapsedTime
    
    rr.shader.fillColor = color(fill())
    if strokeWidth() == 0 then
        rr.shader.strokeColor = rr.shader.fillColor
    else
        rr.shader.strokeColor = color(stroke())
    end

    
    pushMatrix()
    translate(t.x, t.y)
    scale(t.scale or 1)
    rr:draw()
    popMatrix()
    
    -- If we don't yet have a "_lastDrawTime" in __RRects, initialize it:
    if not __RRects._lastDrawTime then
        __RRects._lastDrawTime = ElapsedTime
    end
    
    local now = ElapsedTime
    local dt = now - __RRects._lastDrawTime
    __RRects._lastDrawTime = now
    pruneOldMeshes(__RRects, dt * 1.5)
end

function pruneOldMeshes(cache, threshold)
    local now = ElapsedTime
    for key, rr in pairs(cache) do
        -- skip our special _lastDrawTime entry
        if key ~= "_lastDrawTime" then
            if rr._lastUsed and (now - rr._lastUsed) > threshold then
                cache[key] = nil
            end
        end
    end
end


---------------------------------------------------------
-- 2) IMMEDIATE-MODE BUTTON SYSTEM (with pressing/swapping)
---------------------------------------------------------
---------------------------
-- 2.1) Sensor class created by JMV38
---------------------------
Sensor = class()

function Sensor:init(t)
    self.enabled = true
    self.extra   = t.extra or 0
    self.touches = {}
    self:setParent(t)
    self.events  = {}
    self.doNotInterceptTouches = false
end

function Sensor:setParent(t)
    local p = t.parent or self.parent
    if p.x and p.y and p.w and p.h then
        self.parent = p
    else
        error("Sensor parent must have x,y,w,h coordinates")
    end
    
    self.xywhMode = t.xywhMode or CORNER
    if self.xywhMode == CENTER then
        self.xywh = self.xywhCENTER
    elseif self.xywhMode == CORNER then
        self.xywh = self.xywhCORNER
    elseif self.xywhMode == RADIUS then
        self.xywh = self.xywhRADIUS
    end
end

function Sensor:register(eventName, update, callback)
    table.insert(self.events, {
        name = eventName, 
        callback = callback,
        update   = update
    })
end

-- onTouch: “at least one finger is inside” => state=true
function Sensor:onTouch(callback)
    self:register("onTouch", self.touchUpdate, callback)
end

function Sensor.touchUpdate(event, sensor, t)
    sensor.touching = sensor.touching or {}
    
    -- Track if a finger is inside
    if sensor:inbox(t) and (t.state == BEGAN or t.state == MOVING) then
        sensor.touching[t.id] = true
    elseif (t.state == ENDED or t.state == CANCELLED) then
        sensor.touching[t.id] = nil
    end
    
    -- Check overall “any finger inside?” state
    local newState = false
    for _, inside in pairs(sensor.touching) do
        if inside then
            newState = true
            break
        end
    end
    
    -- If changed, fire callback
    if newState ~= event.state then
        event.state = newState
        event.touch = t
        event:callback()
    end
end

-- onTap: short press with minimal movement
function Sensor:onTap(callback)
    self:register("onTap", self.tapUpdate, callback)
end

local abs = math.abs
function Sensor.tapUpdate(event, sensor, t)
    if t.state == BEGAN and sensor:inbox(t) then
        sensor.touches[t.id] = true
        event.totalMove = 0
        event.t0 = ElapsedTime
        
    elseif t.state == MOVING and sensor.touches[t.id] then
        event.totalMove = (event.totalMove or 0) + abs(t.deltaX) + abs(t.deltaY)
        
    elseif t.state == ENDED and sensor.touches[t.id] then
        sensor.touches[t.id] = nil
        local duration = ElapsedTime - (event.t0 or 0)
        local movement = event.totalMove or 9999
        
        if duration < 0.5 and movement < 20 then
            event.touch = t
            event:callback()
        end
    end
end

function Sensor:touched(t)
    if not self.enabled then return end
    
    if t.state == BEGAN and self:inbox(t) then
        self.touches[t.id] = true
    end
    
    for _, event in ipairs(self.events) do
        event:update(self, t)
    end
    
    local intercepted = self.touches[t.id]
    if self.doNotInterceptTouches then
        intercepted = false
    end
    
    if t.state == ENDED or t.state == CANCELLED then
        self.touches[t.id] = nil
    end
    return intercepted
end

function Sensor:xywhCORNER()
    local p = self.parent
    return p.x, p.y, p.w, p.h
end

function Sensor:xywhCENTER()
    local p = self.parent
    return p.x - p.w/2, p.y - p.h/2, p.w, p.h
end

function Sensor:xywhRADIUS()
    local p = self.parent
    return p.x - p.w, p.y - p.h, p.w*2, p.h*2
end

function Sensor:inbox(t)
    local x,y,w,h = self:xywh()
    return (t.x >= x - self.extra and t.x <= x + w + self.extra
    and t.y >= y - self.extra and t.y <= y + h + self.extra)
end


---------------------------
-- 2.2) Immediate-Mode UI
---------------------------
-- Declare the global toggle. (Change the default as needed.)
parameter.boolean("showCoords", false, function(v) _IMUI.showCoords = v end)

-- We'll keep all internal state in _IMUI
_IMUI = {
    sensor          = nil,  -- We'll create on-demand
    pressedLocation = nil,  -- Where user is currently holding
    lastTapLocation = nil,   -- Where user just tapped
    showCoords = showCoords
}

-- Draw a 10-pixel grid starting from the lower left.
function drawGrid()
    pushStyle()
    strokeWidth(1)
    for i = 0, WIDTH, 10 do
        if math.floor(i/10) % 2 == 0 then
            stroke(255, 100, 100, 220)
        else
            stroke(255, 100, 100, 100)
        end
        line(i, 0, i, HEIGHT)
    end
    
    for j = 0, HEIGHT, 10 do
        if math.floor(j/10) % 2 == 0 then
            stroke(255, 100, 100, 220)
        else
            stroke(255, 100, 100, 100)
        end
        line(0, j, WIDTH, j)
    end
    popStyle()
end

-- Called from touched(touch) in your code
function immediateUI(touch)
    -- If we haven't created the global sensor yet, do so now
    if not _IMUI.sensor then
        _IMUI.sensor = Sensor{
            parent   = { x=0, y=0, w=WIDTH, h=HEIGHT }, 
            xywhMode = CORNER
        }
        
        -- onTouch: set or clear pressedLocation
        _IMUI.sensor:onTouch(function(event)
            if event.state == true then
                -- at least one finger inside
                local t = event.touch
                _IMUI.pressedLocation = vec2(t.x, t.y)
            else
                -- no fingers inside
                _IMUI.pressedLocation = nil
            end
        end)
        
        -- onTap: store the final location
        _IMUI.sensor:onTap(function(event)
            local t = event.touch
            _IMUI.lastTapLocation = vec2(t.x, t.y)
        end)
    end
    
    -- Pass the touch to the sensor
    _IMUI.sensor:touched(touch)
end

-- The immediate-mode button function
function button(label, x, y, w, h, callback, cornerRadius)
    if not w or type(w) == "function" then
        callback = w  -- in this case, the 4th parameter is the callback
        local margin = 45  -- default margin
        local tw, th = textSize(label)
        w = tw + margin
        h = th + margin * 0.5
    end
    
    -- Reset grid flag once per frame
    local currentFrame = math.floor(ElapsedTime * 1000)
    if _IMUI.lastGridFrame ~= currentFrame then
        _IMUI.gridDrawn = false
        _IMUI.lastGridFrame = currentFrame
    end
    
    if _IMUI.showCoords and not _IMUI.gridDrawn then
        drawGrid()
        _IMUI.gridDrawn = true
    end
    
    -- If there's no sensor yet, create it now (same logic as in immediateUI)
    if not _IMUI.sensor then
        _IMUI.sensor = Sensor{
            parent   = { x=0, y=0, w=WIDTH, h=HEIGHT }, 
            xywhMode = CORNER
        }
        
        _IMUI.sensor:onTouch(function(event)
            if event.state == true then
                local t = event.touch
                _IMUI.pressedLocation = vec2(t.x, t.y)
            else
                _IMUI.pressedLocation = nil
            end
        end)
        
        _IMUI.sensor:onTap(function(event)
            local t = event.touch
            _IMUI.lastTapLocation = vec2(t.x, t.y)
        end)
    end
    
    local pressLoc = _IMUI.pressedLocation
    local tapLoc   = _IMUI.lastTapLocation
    
    -- Check if pressed
    local isPressed = false
    if pressLoc and pointInRectByCurrentMode(pressLoc.x, pressLoc.y, x, y, w, h) then
        isPressed = true
    end
    
    -- Check if tapped
    if tapLoc and pointInRectByCurrentMode(tapLoc.x, tapLoc.y, x, y, w, h) then
        if callback then callback() end
        -- Clear it so we don’t double-trigger multiple buttons with one tap
        _IMUI.lastTapLocation = nil
    end
    
    -- Now draw
    
    pushStyle()
    
    local originalFill   = color(fill())   -- whatever the user has set as fill()
    local originalStroke = color(stroke()) -- whatever the user has set as stroke()
    local textColor 
    
    -- If showCoords is enabled, make the fill semitransparent.
    if _IMUI.showCoords then
        originalFill = color(originalFill.r, originalFill.g, originalFill.b, 100)
    end
    
    -- If isPressed, swap fill and stroke so it "inverts" when pressed
    if isPressed then
        fill(originalStroke)
        stroke(originalFill)
        textColor = originalFill
    else
        fill(originalFill)
        stroke(originalStroke)
        textColor = originalStroke
    end
    
    -- Draw the rounded rectangle (corners=15 means all corners rounded)
    roundedRectangle{
        x = x,
        y = y,
        w = w,
        h = h,
        corners = 15,
        radius = cornerRadius or 18    -- adjust corner radius as you like
    }
    
    local mode = rectMode()
    local textX, textY
    if mode == CORNER then
        textX = x + w/2
        textY = y + h/2
    elseif mode == CENTER or mode == RADIUS then
        textX = x
        textY = y
    else
        textX = x + w/2
        textY = y + h/2
    end
    
    -- Use the "stroke" color for text
    fill(textColor)
    textMode(CENTER)
    textAlign(CENTER)
    text(label, textX, textY)
    
    popStyle()
    
    -- If showCoords is enabled, overlay the button's coordinates.
    if _IMUI.showCoords then
        pushStyle()
        font("AmericanTypewriter-Bold")
        fontSize(16)
        fill(255)
        local overlayX, overlayY
        if rectMode() == CORNER then
            overlayX = x + w/2
            overlayY = y + h/2
        else
            overlayX = x
            overlayY = y
        end
        text(string.format("(%d, %d)", math.floor(x), math.floor(y)), overlayX, overlayY)
        popStyle()
    end
end

-- A helper for geometry
function pointInRectByCurrentMode(px, py, cx, cy, w, h)
    if false then -- set to true for debugging
        pushStyle()
        noFill()
        stroke(193, 236, 67)
        strokeWidth(8)
        rect(cx, cy, w, h)
        popStyle()
    end
    
    local mode = rectMode()
    local left, right, bottom, top
    if mode == CORNER then
        left   = cx
        bottom = cy
        right  = cx + w
        top    = cy + h
    elseif mode == CENTER then
        left   = cx - w/2
        bottom = cy - h/2
        right  = cx + w/2
        top    = cy + h/2
    elseif mode == RADIUS then
        left   = cx - w
        bottom = cy - h
        right  = cx + w
        top    = cy + h
    else
        -- default to center mode
        left   = cx - w/2
        bottom = cy - h/2
        right  = cx + w/2
        top    = cy + h/2
    end
    
    return (px >= left and px <= right and py >= bottom and py <= top)
end
